==> wordle-1.0.tar.gz: wordle/solve.c <==
/*
    This file is part of Wordle Solver.
    Copyright (C) 2022  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "dict.h"
#include "lookup.h"
#include "tree.h"
#include "builtin_dict.h"
#include "builtin_tree.h"
#include "solve.h"

// ****************************************************************************
// 
// ****************************************************************************
void solver_init(solver *s, int flags)
{
	s->flags = flags;
	dict_init(&s->d);
	dict_builtin_full(&s->d);
	tree_builtin_get(&s->size, &s->root, &s->node_buffer, flags);
	s->current = s->root;
}

void solver_reset(solver *s)
{
	s->current = s->root;
}

void solver_clear(solver *s)
{
	dict_clear(&s->d);
}

// ****************************************************************************
// 
// ****************************************************************************
node_index solver_advise(solver *s)
{
	node *nb = s->node_buffer;
	
	return nb[s->current].guess;
}

int solver_next(solver *s, hash_index h)
{
	node *nb = s->node_buffer;
	node_index i = nb[s->current].child;
	
	while (i != NODE_NONE) {
		if ((nb[i].hash == HASH_ANY)
		||  (nb[i].hash == h)) {
			s->current = i;
			return 0;
		}
		
		i = nb[i].sibling;
	}
	
	return -1;
}

// ****************************************************************************
// Solve
// ****************************************************************************
void solver_interactive(solver *s)
{
	size_t depth = 0;
	
	while (1) {
		node_index ti = solver_advise(s);
		
		word guess = s->d.words[ti];
		
		while (1) {
			printf(
				"\nPlay %2zd>        %s\n\n"
				"Answer [Gg green, Yy yellow, /. gray] "
				"or [R restart, Q quit]\n\n"
				"    | ",
				depth, word_str(guess).s);
			fflush(stdout);
			
			char buff[LENGTH + 3];
			buff[LENGTH + 2] = 0;
			
			char *a = fgets(buff, LENGTH + 2, stdin);
			
			if (!a) {
				printf("    Stop\n\n");
				return;
			}
		
			size_t l = strlen(a);
			
			if ((l > 0) && (a[l - 1] == '\n'))
				a[--l] = 0;
			
			if ((strcmp(a, "r") == 0) || (strcmp(a, "R") == 0)) {
				printf("\n    *** restart ***\n\n");
				solver_reset(s);
				depth = 0;
				break;
			}
			
			if ((strcmp(a, "q") == 0) || (strcmp(a, "Q") == 0)) {
				printf("    Stop\n\n");
				return;
			}

			if (l != LENGTH) {
				printf("    Invalid length '%s'\n", a);
				continue;
			}
			
			int valid = 1;
			word aw = word_build(a);
			
			for (size_t j = 0; j < LENGTH; j++) {
				if ((aw.s[j] != 'G') && (aw.s[j] != 'g')
				&&  (aw.s[j] != 'Y') && (aw.s[j] != 'y')
				&&  (aw.s[j] != '.') && (aw.s[j] != '/')
				&&  (aw.s[j] != ' ') && (aw.s[j] != '-'))
					valid = 0;
			}
			
			if (!valid) {
				printf("    Invalid [GY.] '%s'\n", a);
			}
			
			hash_index hash = lut_hash_parse(aw);
			
			if (solver_next(s, hash)) {
				printf("    !!!! Out of advice !!!!"
					"\n");
				depth--;
			}
			
			break;
		}

		depth++;
	}
}

// ****************************************************************************
// 
// ****************************************************************************
//#define PMAX(...)	fprintf(stderr, __VA_ARGS__)
#define PMAX(...)
void solver_all(solver *s)
{
	fprintf(stderr, "Building lookup table...\n");
	lut *t = malloc(sizeof(lut));
	if (!t) {
		fprintf(stderr, "Could not allocate lookup table.\n");
		exit(1);
	}
	
	lut_build(t, &s->d);

	dict sol;
	dict_init(&sol);
	
	if (s->flags & GAME_CHEAT) {
		dict_builtin_solutions(&sol);
	}
	
	fprintf(stderr, "Playing...\n");
	
	word *w = s->d.words;
	size_t success = 0;
	size_t fail = 0;
	size_t depth_sum = 0;
	size_t depth_max = 0;
	size_t depth_max_count = 0;

	for (dict_index target = 0; target < DICT_SIZE; target++) {
		if (s->flags & GAME_CHEAT) {
			if (!dict_has(&sol, w[target]))
				continue;
		}
		
		solver_reset(s);
		size_t depth = 0;
		
		while (1) {
			dict_index guess = solver_advise(s);
			
			if (guess == target) {
				if (depth_max < depth) {
					PMAX("  depth %2zd: %5zd: '%s'\n",
						depth, 0l,
						word_str(w[target]).s);
					depth_max = depth;
					depth_max_count = 1;
				} else if (depth_max == depth) {
					depth_max_count++;
					PMAX("  depth %2zd: %5zd: '%s'\n",
						depth, depth_max_count,
						word_str(w[target]).s);
				}
				depth_sum += depth;
				success++;
				break;
			}

			hash_index h = lut_lookup(t, guess, target);
			
			if (solver_next(s, h)) {
				fprintf(stderr, "Solver ran out of advice "
					"for target %d '%s' at depth %zd\n",
					target,
					word_str(w[target]).s,
					depth);
				fail++;
				break;
			}

			depth++;
		}
	}
	
	double depth_avg = (double)depth_sum / success;
	
	fprintf(stderr,
		"   Success:   %5zd\n"
		"      Fail:   %5zd\n"
		" Max depth:   %5zd     [win in %5zd     plays]"
			" (%5zd / %5zd)\n"
		" Avg depth:   %9.3f [win in %9.3f plays]\n",
		success, fail,
		depth_max, depth_max + 1, depth_max_count, success,
		depth_avg, 1.0 + depth_avg);

	dict_clear(&sol);
}


// ****************************************************************************
// 
// ****************************************************************************
static node_index solver_connected(solver *s,
	uint8_t *swept, depth_index depth, node_index p)
{
	node *nb = s->node_buffer;
	node_index count = 1;
	node_index i = nb[p].child;
	
	if (nb[p].depth != depth) {
		fprintf(stderr, "node %zd has depth %u but claims %u\n",
			p, depth, nb[p].depth);
		exit(1);
	}
	
	swept[p]++;
	
	while (i != NODE_NONE) {
		count += solver_connected(s, swept, depth + 1, i);
		i = nb[i].sibling;
	}
	
	return count;
}


int solver_connectivity(solver *s)
{
	uint8_t *swept = malloc(s->size);
	if (!swept) {
		fprintf(stderr, "Cannot allocate %zd-bytes swept vector\n",
			s->size);
		exit(1);
	}
	
	memset(swept, 0, s->size);

	node_index connected = solver_connected(s, swept, 0, s->root);
	
	fprintf(stderr,
		" Connected:   %5zd / %5zd (depths ok)\n",
		connected, s->size);
	
	node *nb = s->node_buffer;
	word *w = s->d.words;
	int err = (connected != s->size);

	for (node_index i = 0; i < s->size; i++) {
		if (swept[i] == 1)
			continue;

		dict_index guess = nb[i].guess;
		
		fprintf(stderr,
			"node %5zd guess %5u '%s' depth %u: swept = %u, "
			"sibling: %5zd, child: %5zd\n",
			i, guess, word_str(w[guess]).s, nb[i].depth, swept[i],
			nb[i].sibling, nb[i].child);
		err = 1;
	}
	
	free(swept);
	
	return err;
}

// ****************************************************************************
// 
// ****************************************************************************
void solver_export_js(int flags)
{
	dict d;
	dict_init(&d);
	dict_builtin_full(&d);
	
	node_index count;
	node_index root;
	node *nodes;
	char name[3] = { 'a', 'r', 0 };
	
	if (flags & GAME_CHEAT)
		name[0] = 'c';
	if (flags & GAME_HARD)
		name[1] = 'h';
	
	tree_builtin_get(&count, &root, &nodes, flags);
	
	printf("var tree_%s_root = %zd;\n", name, root);
	printf("var tree_%s_nodes = [\n", name);
	
	for (node_index i = 0; i < count; i++) {
		printf("{a:'%s',g:'%s',c:%zd,s:%zd}%s\n",
			word_str(lut_hash_word(nodes[i].hash)).s,
			word_str(d.words[nodes[i].guess]).s,
			nodes[i].child, nodes[i].sibling,
			(i == count - 1) ? "" : ",");
	}
	printf("];\n");
	
	dict_clear(&d);
}


static void solver_subtree_txt(dict *d, node *nodes, node_index p)
{
	printf("%*s%d %s %s\n",
		nodes[p].depth * 2, "",
		nodes[p].depth,
		word_str(lut_hash_word(nodes[p].hash)).s,
		word_str(d->words[nodes[p].guess]).s);
	
	node_index i = nodes[p].child;
	
	while (i != NODE_NONE) {
		solver_subtree_txt(d, nodes, i);
		
		i = nodes[i].sibling;
	}
}

void solver_export_txt(int flags)
{
	dict d;
	dict_init(&d);
	dict_builtin_full(&d);

	node_index count;
	node_index root;
	node *nodes;
	
	tree_builtin_get(&count, &root, &nodes, flags);
	
	solver_subtree_txt(&d, nodes, root);
	
	dict_clear(&d);
}



// ****************************************************************************
// 
// ****************************************************************************

==> wordle-2.0.tar.gz: wordle/solve.c <==
/*
    This file is part of Wordle Solver.
    Copyright (C) 2022  Laurent Poirrier

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as published
    by the Free Software Foundation, version 3 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include "check.h"
#include "dict.h"
#include "lookup.h"
#include "tree.h"
#include "builtin_dict.h"
#include "builtin_tree.h"
#include "solve.h"

// ****************************************************************************
// 
// ****************************************************************************
void solver_init(solver *s, int flags)
{
	s->flags = flags;
	dict_init(&s->d);
	dict_builtin_full(&s->d);
	tree_builtin_get(&s->size, &s->root, &s->node_buffer, flags);
	s->current = s->root;
}

void solver_reset(solver *s)
{
	s->current = s->root;
}

void solver_clear(solver *s)
{
	dict_clear(&s->d);
}

// ****************************************************************************
// 
// ****************************************************************************
node_index solver_advise(solver *s)
{
	node *nb = s->node_buffer;
	
	return nb[s->current].guess;
}

int solver_next(solver *s, hash_index h)
{
	node *nb = s->node_buffer;
	node_index i = nb[s->current].child;
	
	while (i != NODE_NONE) {
		if ((nb[i].hash == HASH_ANY)
		||  (nb[i].hash == h)) {
			s->current = i;
			return 0;
		}
		
		i = nb[i].sibling;
	}
	
	return -1;
}

// ****************************************************************************
// Solve
// ****************************************************************************
void solver_interactive(solver *s)
{
	size_t depth = 0;
	
	while (1) {
		node_index ti = solver_advise(s);
		
		word guess = s->d.words[ti];
		
		while (1) {
			printf(
				"\nPlay %2zd>        %s\n\n"
				"Answer [Gg green, Yy yellow, /. gray] "
				"or [R restart, Q quit]\n\n"
				"    | ",
				depth, word_str(guess).s);
			fflush(stdout);
			
			char buff[LENGTH + 3];
			buff[LENGTH + 2] = 0;
			
			char *a = fgets(buff, LENGTH + 2, stdin);
			
			if (!a) {
				printf("    Stop\n\n");
				return;
			}
		
			size_t l = strlen(a);
			
			if ((l > 0) && (a[l - 1] == '\n'))
				a[--l] = 0;
			
			if ((strcmp(a, "r") == 0) || (strcmp(a, "R") == 0)) {
				printf("\n    *** restart ***\n\n");
				solver_reset(s);
				depth = 0;
				break;
			}
			
			if ((strcmp(a, "q") == 0) || (strcmp(a, "Q") == 0)) {
				printf("    Stop\n\n");
				return;
			}

			if (l != LENGTH) {
				printf("    Invalid length '%s'\n", a);
				continue;
			}
			
			int valid = 1;
			word aw = word_build(a);
			
			for (size_t j = 0; j < LENGTH; j++) {
				if ((aw.s[j] != 'G') && (aw.s[j] != 'g')
				&&  (aw.s[j] != 'Y') && (aw.s[j] != 'y')
				&&  (aw.s[j] != '.') && (aw.s[j] != '/')
				&&  (aw.s[j] != ' ') && (aw.s[j] != '-'))
					valid = 0;
			}
			
			if (!valid) {
				printf("    Invalid [GY.] '%s'\n", a);
			}
			
			hash_index hash = lut_hash_parse(aw);
			
			if (solver_next(s, hash)) {
				printf("    !!!! Out of advice !!!!"
					"\n");
				depth--;
			}
			
			break;
		}

		depth++;
	}
}

// ****************************************************************************
// 
// ****************************************************************************
//#define PMAX(...)	fprintf(stderr, __VA_ARGS__)
#define PMAX(...)
void solver_all(solver *s)
{
	fprintf(stderr, "Building lookup table...\n");
	lut *t = malloc(sizeof(lut));
	if (!t) {
		fprintf(stderr, "Could not allocate lookup table.\n");
		exit(1);
	}
	
	lut_build(t, &s->d);

	dict sol;
	dict_init(&sol);
	
	if (s->flags & GAME_CHEAT) {
		dict_builtin_solutions(&sol);
	} else if (s->flags & GAME_ILL) {
		dict_builtin_ill(&sol);
	}
	
	fprintf(stderr, "Playing...\n");
	
	word *w = s->d.words;
	size_t success = 0;
	size_t fail = 0;
	size_t depth_sum = 0;
	size_t depth_max = 0;
	size_t depth_max_count = 0;

	for (dict_index target = 0; target < DICT_SIZE; target++) {
		if (s->flags & (GAME_CHEAT | GAME_ILL)) {
			if (!dict_has(&sol, w[target]))
				continue;
		}
		
		solver_reset(s);
		size_t depth = 0;
		
		while (1) {
			dict_index guess = solver_advise(s);
			
			if (guess == target) {
				if (depth_max < depth) {
					PMAX("  depth %2zd: %5zd: '%s'\n",
						depth, 0l,
						word_str(w[target]).s);
					depth_max = depth;
					depth_max_count = 1;
				} else if (depth_max == depth) {
					depth_max_count++;
					PMAX("  depth %2zd: %5zd: '%s'\n",
						depth, depth_max_count,
						word_str(w[target]).s);
				}
				depth_sum += depth;
				success++;
				break;
			}

			hash_index h = lut_lookup(t, guess, target);
			
			if (solver_next(s, h)) {
				fprintf(stderr, "Solver ran out of advice "
					"for target %d '%s' at depth %zd\n",
					target,
					word_str(w[target]).s,
					depth);
				fail++;
				break;
			}

			depth++;
		}
	}
	
	double depth_avg = (double)depth_sum / success;
	
	fprintf(stderr,
		"   Success:   %5zd\n"
		"      Fail:   %5zd\n"
		" Max depth:   %5zd      [win in %5zd      guesses]"
			" (%5zd / %5zd)\n"
		" Avg depth:   %10.4f [win in %10.4f guesses]\n"
		"       Sum:   %5zd      [win in %5zd      guesses]\n",
		success, fail,
		depth_max, depth_max + 1, depth_max_count, success,
		depth_avg, 1.0 + depth_avg,
		depth_sum, depth_sum + success);

	dict_clear(&sol);
	free(t);
}


// ****************************************************************************
// 
// ****************************************************************************
static node_index solver_connected(solver *s,
	uint8_t *swept, depth_index depth, node_index p)
{
	node *nb = s->node_buffer;
	node_index count = 1;
	node_index i = nb[p].child;
	
	if (nb[p].depth != depth) {
		fprintf(stderr, "node %zd has depth %u but claims %u\n",
			p, depth, nb[p].depth);
		exit(1);
	}
	
	swept[p]++;
	
	while (i != NODE_NONE) {
		count += solver_connected(s, swept, depth + 1, i);
		i = nb[i].sibling;
	}
	
	return count;
}


int solver_connectivity(solver *s)
{
	uint8_t *swept = malloc(s->size);
	if (!swept) {
		fprintf(stderr, "Cannot allocate %zd-bytes swept vector\n",
			s->size);
		exit(1);
	}
	
	memset(swept, 0, s->size);

	node_index connected = solver_connected(s, swept, 0, s->root);
	
	fprintf(stderr,
		" Connected:   %5zd / %5zd (depths ok)\n",
		connected, s->size);
	
	node *nb = s->node_buffer;
	word *w = s->d.words;
	int err = (connected != s->size);

	for (node_index i = 0; i < s->size; i++) {
		if (swept[i] == 1)
			continue;

		dict_index guess = nb[i].guess;
		
		fprintf(stderr,
			"node %5zd guess %5u '%s' depth %u: swept = %u, "
			"sibling: %5zd, child: %5zd\n",
			i, guess, word_str(w[guess]).s, nb[i].depth, swept[i],
			nb[i].sibling, nb[i].child);
		err = 1;
	}
	
	free(swept);
	
	return err;
}

// ****************************************************************************
// 
// ****************************************************************************
void solver_export_js(int flags)
{
	dict d;
	dict_init(&d);
	dict_builtin_full(&d);
	
	node_index count;
	node_index root;
	node *nodes;
	char name[3] = { 'a', 'r', 0 };
	
	if (flags & GAME_ILL)
		name[0] = 'i';
	else if (flags & GAME_CHEAT)
		name[0] = 'c';
	
	if (flags & GAME_HARD)
		name[1] = 'h';
	
	tree_builtin_get(&count, &root, &nodes, flags);
	
	printf("var tree_%s_root = %zd;\n", name, root);
	printf("var tree_%s_nodes = [\n", name);
	
	for (node_index i = 0; i < count; i++) {
		printf("{a:'%s',g:'%s',c:%zd,s:%zd}%s\n",
			word_str(lut_hash_word(nodes[i].hash)).s,
			word_str(d.words[nodes[i].guess]).s,
			nodes[i].child, nodes[i].sibling,
			(i == count - 1) ? "" : ",");
	}
	printf("];\n");
	
	dict_clear(&d);
}


// ****************************************************************************
// 
// ****************************************************************************
static void solver_subtree_txt(dict *d, node *nodes, node_index p)
{
	printf("%*s%d %s %s\n",
		nodes[p].depth * 2, "",
		nodes[p].depth,
		word_str(lut_hash_word(nodes[p].hash)).s,
		word_str(d->words[nodes[p].guess]).s);
	
	node_index i = nodes[p].child;
	
	while (i != NODE_NONE) {
		solver_subtree_txt(d, nodes, i);
		
		i = nodes[i].sibling;
	}
}

void solver_export_txt(int flags)
{
	dict d;
	dict_init(&d);
	dict_builtin_full(&d);

	node_index count;
	node_index root;
	node *nodes;
	
	tree_builtin_get(&count, &root, &nodes, flags);
	
	solver_subtree_txt(&d, nodes, root);
	
	dict_clear(&d);
}

// ****************************************************************************
// 
// ****************************************************************************
void solver_path_csv(dict *d, lut *t,
	node_index root, node *nodes, dict_index x)
{
	node_index i = root;
	
	while (1) {
		dict_index w = nodes[i].guess;
		
		printf("%s", word_str(d->words[w]).s);

		if (w == x)
			break;
		
		hash_index h = lut_lookup(t, w, x);
		
		node_index j = nodes[i].child;
		
		while (1) {
			if (j == NODE_NONE) {
				fprintf(stderr, "Out of guesses for %s!\n",
					word_str(d->words[x]).s);
				exit(1);
			}
			
			if (nodes[j].hash == HASH_ANY)
				break;
			
			if (nodes[j].hash == h)
				break;
			
			j = nodes[j].sibling;
		}
		
		i = j;
		printf(",");
	}
	
	printf("\n");
}

void solver_export_csv(int flags)
{
	dict d;
	dict_init(&d);
	dict_builtin_full(&d);

	dict x;
	dict_init(&x);
	
	if (flags & GAME_ILL) {
		dict_builtin_ill(&x);
	} else if (flags & GAME_CHEAT) {
		dict_builtin_solutions(&x);
	} else {
		dict_builtin_full(&x);
	}
	
	fprintf(stderr, "Building lookup table...\n");
	lut *t = malloc(sizeof(lut));
	if (!t) {
		fprintf(stderr, "Could not allocate lookup table.\n");
		exit(1);
	}
	
	lut_build(t, &d);

	fprintf(stderr, "Playing...\n");

	node_index count;
	node_index root;
	node *nodes;
	
	tree_builtin_get(&count, &root, &nodes, flags);
	
	for (dict_index i = 0; i < d.count; i++) {
		if (!dict_has(&x, d.words[i]))
			continue;
		
		solver_path_csv(&d, t, root, nodes, i);
	}
	
	dict_clear(&x);
	dict_clear(&d);
	free(t);
}

// ****************************************************************************
// 
// ****************************************************************************
#define NODES_MAX	65536
#define DEPTH_MAX	32

void solver_import_txt(void)
{
	node *nodes = malloc(NODES_MAX * sizeof(node));
	
	if (!nodes) {
		fprintf(stderr, "malloc failed\n");
		exit(1);
	}
	
	dict d;
	dict_init(&d);
	dict_builtin_full(&d);

	node_index cur[DEPTH_MAX];
	depth_index d0 = DEPTH_INF;
	node_index n = 0;
	
	while (1) {
		int d1_i;
		char a1[6];
		char g1[6];
		
		if (scanf(" %d %5s %5s\n", &d1_i, a1, g1) != 3)
			break;
		
		check(d1_i, >=, 0);
		check(d1_i, <, DEPTH_MAX);
		
		depth_index d1 = d1_i;
		hash_index h = lut_hash_parse(word_build(a1));
		dict_index w = dict_index_get(&d, word_build(g1));
		
		check(w, !=, DICT_NONE);

		nodes[n].depth = d1;
		nodes[n].hash = h;
		nodes[n].guess = w;
		nodes[n].child = NODE_NONE;
		nodes[n].sibling = NODE_NONE;
		
		if (d0 == DEPTH_INF) {
			check(d1, ==, 0);
			check(n, ==, 0);
		} else {
			check(d1, >=, 1);
			check(d1, <=, d0 + 1);
			
			if (d1 > d0) {
				check(nodes[cur[d0]].child, ==, NODE_NONE);
				nodes[cur[d0]].child = n;
			} else {
				nodes[cur[d1]].sibling = n;
			}
		}

		cur[d1] = n;
		d0 = d1;
		n++;
	}
	
	printf("#define TREE_ROOT 0\n");
	printf("#define TREE_SIZE %zd\n", n);
	printf("static node tree_nodes[TREE_SIZE] = {\n");
	
	for (node_index i = 0; i < n; i++) {
		printf("{ %u, %u, %u, %zd, %zd }, // %zd %s\n",
			nodes[i].hash, nodes[i].depth,
			nodes[i].guess, nodes[i].child, nodes[i].sibling,
			i, word_str(d.words[nodes[i].guess]).s);
	}
	
	printf("};\n");
	
	dict_clear(&d);
	free(nodes);
}

// ****************************************************************************
// 
// ****************************************************************************
